"use strict";
this.name           = "ils";
this.author         = "Norby";
this.copyright      = "2016 Norby";
this.description    = "Instrument Landing System";
this.licence        = "CC BY-NC-SA 4.0";
this.version        = "1.14"

//internal properties, should not touch
this.$F = null; //FCB pointer
this.$L = null; //last ILS user ship
this.$N = null; //wide NPC in last stage
this.$P = null; //previous heading in FCB
this.$S = null; //the target station
this.$S_max_speed = -1
this.$S_max_pitch = -1
this.$S_max_roll = -1
this.$S_max_yaw = -1
this.$T = null; //Towbar worldScript pointer
this.$Targets = []; //Result of scanner check
this.$Timer = null; //Timer for scanner check
this.$V = null; //visualEffect for the white flashing marker
this.$PI = Math.PI;
this.$maxSpeedRef = 900; // reference for max docking speed: bellow that we try to moderate rotation for visual effects
this.$counter = 0;
this.$logging = true;
this.$debug = false; // turns on _a lot_ of tracing messages to the log

//
//worldscript events
//

//-----------------------------------------------------------------------------------
this.startUp = function _ils_startUp() {
    this.$T = worldScripts.towbar;
    this.ship = player.ship;
    if (!isValidFrameCallback(this.$F)) this.$F = addFrameCallback(this.$FCB.bind(this));
    this.$Timer = new Timer(this, this.$Timed.bind(this), 1, 1);//scanner check in every seconds
}

//-----------------------------------------------------------------------------------
this.equipmentDamaged = this.equipmentDestroyed = function _ils_equipmentDamaged(equipment) {
    if (equipment === "EQ_ILS") this.$ClearILS(); //stop ILS if working
}

/* another oxp cause a bug which turn off ILS immediately in the next lines so do not use this event
this.shipTargetLost = function() {
    log(this.name, "ILS OFF 1");//debug
    this.$ClearILS();
}
*/

//-----------------------------------------------------------------------------------
this.shipWillDockWithStation = this.shipWillEnterWitchspace = this.shipDied = function _ils_stopILS() {
    this.$ClearILS(true);//nomsg
}

//-----------------------------------------------------------------------------------
this.shipWillLaunchFromStation = function _ils_shipWillLaunchFromStation(station) {
    if (this.$L && this.$L != player.ship && this.$L.isValid && this.$L.isInSpace)
        //stop landing NPC when player launch
        this.$L.desiredSpeed = 0; //stop landing NPC when player launch
    this.$ClearILS(true);
    this.ship = player.ship;
}


//
//ILS methods
//


//-----------------------------------------------------------------------------------
this.$ClearILS = function _ils_ClearILS(nomsg) {
    if (this.$S && this.$S_max_speed > 0) {
        var S = this.$S
        if (this.$logging) log(this.name, "Restoring "+S.displayName+" max speed back to "+this.$S_max_speed)
        S.maxSpeed = this.$S_max_speed
        S.maxRoll = this.$S_max_roll
        S.maxPitch = this.$S_max_pitch
        S.maxYaw = this.$S_max_yaw
        this.$S_max_speed = -1
        this.$S_max_roll = -1
        this.$S_max_pitch = -1
        this.$S_max_yaw = -1
    }
    this.$N = null;
    this.$P = null;
    this.$S = null;
    if (this.$V && this.$V.isVisualEffect) {
        this.$V.remove();
        if (!nomsg && this.$L == player.ship)
            player.consoleMessage("ILS off");
    }
    this.$V = null;
    this.$L = null;
//    if( isValidFrameCallback( this.$F ) ) removeFrameCallback( this.$F );
//    log(this.name, "ILS OFF");//debug
}

//-----------------------------------------------------------------------------------
this.$Timed = function _ils_Timed() {
    var that = _ils_Timed;
    var targets = (that.targets = that.targets || this.$Targets);
    var pship = this.ship;  // player ship
    var t = (pship ? pship.target : null);
    var station = this.$S;
    var mainStation = system.mainStation;
    var ref_station;

    function _append_to_targets(value) {
        targets.push(value);
    }

    if (t && t.isValid && t.isStation && t.dataKey.indexOf("carrier") == -1)
        // player is targetting a station, use it
        ref_station = t;
    else if (station && station.isValid) {
        // last station ILS was used is set and valid
        if (mainStation && pship.position.distanceTo(station) > 100000)
            // mainStation is closer
            ref_station = mainStation;
        else
            ref_station = station;
    } else if (mainStation)
        ref_station = mainStation;

    if (!ref_station || !ref_station.isValid) return;

    var old_targets_len = (targets ? targets.length : 0);
    var new_targets = ref_station.checkScanner(true);   //much faster than filteredEntities
    if (new_targets) new_targets.forEach(_append_to_targets);
    if (old_targets_len) targets.splice(0, old_targets_len);
}

//-----------------------------------------------------------------------------------
this.$NPC = function _ils_NPC(t) { //target is a landing NPC ship?
    if (t && t.isPiloted && t.maxSpeed > 0 && t != player.ship
        && ( t.AI == "dockingAI.plist" || t.dockingInstructions )
        && t.target && t.target.isStation && t.target.maxSpeed === 0) {
        return(true);
//                 //wide NPC need rolling target station to turn on ILS - disabled since can roll
//                 if( t.target.roll != 0 ) return(true);
//                 var u = t.target.subEntities;
//                 var i = 0;
//                 if( u ) while( u[i] && !u[i].isDock ) i++;
//                 if( u && u[i] && u[i].isDock &&
//                         t.boundingBox.x > Math.min( u[i].boundingBox.x, u[i].boundingBox.y ) ) {
//                 return(true); //not wide so fit into without roll
//                 }
    }
    return(false);
}

//-----------------------------------------------------------------------------------
this.$FCB = function _ils_FCB(delta) {
    var that = _ils_FCB;
//    var w = worldScripts.ils;
//    player.consoleMessage(this.$S+" "+delta); //debug
    var pship = this.ship;
    var p = pship;
    if( !p || !p.isValid ) return;
    var t = p.target;
    // de-referencing for performance
    var PI = (that.PI = that.PI || this.$PI);
    var halfPI = (that.halfPI = that.halfPI || PI / 2);
    var counter = (that.counter = that.counter || this.$counter);
    var debug = (that.debug = that.debug || this.$debug);
    var logging = (that.logging = that.logging || this.$logging);
    var L = this.$L; 

    var pp = p.position;
    var tis = false; //target is station but not a carrier?
    var npc = false; //there is a landing NPC ship?
    var near = 10000; //NPCs use ILS within this distance
    var noinjdist = 2000; //switch off injectors within this distance
    var maxw = 64; //safe max. width of ship to fit into the dock height without roll

    if (debug) {
        // the counter is used to limit the tracing messages to a fraction of the FB calls
        counter++;
        counter = counter % 200;
        this.$counter = counter;
    }
//    if (debug) log(this.name, "FCB starts, delta:"+delta.toFixed(5)+", last ILS user: "+(L ? L.displayName : L)+", player target: "+(t ? t.displayName : t)+", last station: "+this.$S);

    if (t && (t.isStation && t.dataKey.indexOf("carrier") == -1 || t == L)
        && !p.missilesOnline //not in missile targeting mode
        && (p.alertCondition !== 3 || !p.weaponsOnline) //not in red alert with online weapons
        && p.equipmentStatus("EQ_ILS") === "EQUIPMENT_OK" ) {
        // ILS is available to player and player attempting to use it:
        //      - player's target is the station or the current ILS user
        if (!L || L == p || !L.isValid || !L.isInSpace) { 
            // ILS is free to use by the player:
            //      - no last ILS user or
            //      - last ILS user is the player or
            //      - last ILS user, a NPC, is no more or already docked
            if (t != L) {
                if (debug && counter == 0) log(this.name, "ILS free");
                tis = true;
            } else if (L.target && L.target.isStation) {
                if (debug) log(this.name, "current ILS user already docked");
                tis = true; //set target to the station when npc is docked
                if (p.target != L.target) {
                    if (debug) log(this.name, "setting "+p.displayName+" target to "+L.target.displayName);
                    t = L.target; 
                    if (player.dockingClearanceStatus !== "DOCKING_CLEARANCE_STATUS_NONE" && p.target != t) p.target = t;
                }
            }
        } 
          else 
            // ILS is being used by a NPC
            if (player.dockingClearanceStatus !== "DOCKING_CLEARANCE_STATUS_NONE" && p.target != L && pp.distanceTo(L) < p.scannerRange) {
                //prevent repeated messages when target is out of range
                if (p.target != L) {
                    if (debug) log(this.name, "setting "+p.displayName+" target to last ILS user "+L.displayName);
                    p.target = L;
                }
                player.consoleMessage("ILS locked by "+L.displayName, 5);
            }
    }

    if (p.isInSpace && (player.dockingClearanceStatus === "DOCKING_CLEARANCE_STATUS_GRANTED" 
        || player.dockingClearanceStatus === "DOCKING_CLEARANCE_STATUS_TIMING_OUT")) {
        // player has docking clearance
        if (debug && counter == 0) log(this.name, "player has docking clearance, target: ", (t ? t.displayName : t)+(L ? ", ILS user:"+L.displayName+" inSpace:"+L.isInSpace : ", ILS free for use"));
        if (L && L != p && L.isValid) { 
            //stop NPC when player has docking clearance
            L.desiredSpeed = 0;
            if (p.target != L.target) p.target = L.target;
        }
    } else if (L && L != p && L.isValid && L.isInSpace && this.$NPC(L)) { // player hasn't docking clearance
        // current user is a NPC who is still docking
        npc = true;
        t = L;
    } else if (!p.isInSpace
               || player.dockingClearanceStatus === "DOCKING_CLEARANCE_STATUS_NONE" // player hasn't docking clearance and ILS is free
               || player.dockingClearanceStatus === "DOCKING_CLEARANCE_STATUS_REQUESTED"
               || !tis) { 
        // a NPC can use ILS when it's free and the player can't use it:
        //      - player hasn't requested docking clearance
        //      - player is waiting for docking clearance
        //      - ILS isn't available to player
        //      - player is docked
        t = null;
        var targets = this.$Targets; //filled by checkScanner() in timer
        if (targets && targets.length > 0) {
            var bestsd = 0; //best speed/distance mean fastest landing so far
            for (var i = 0; i < targets.length; i++) {
                var n = targets[i];
                if (this.$NPC(n)) {
                    // target is a landing NPC
                    // ILS on the ship which probably arrive first
                    var nd = n.position.distanceTo(n);
                    var sd = n.maxSpeed/nd;
                    if (nd < near && nd > noinjdist && n.fuel > 0.2
                        && n.equipmentStatus("EQ_FUEL_INJECTION") === "EQUIPMENT_OK" ) {
                        if (n.injectorSpeedFactor > 0)
                            sd = n.injectorSpeedFactor * sd;
                        else 
                            sd = 7 * sd; //for Oolite 1.80
                    }
                    if (nd < near 
                        && (!t || sd > bestsd) //first or faster
                        && n.boundingBox.x < 210) {//exclude bigships
                        // replace best candidate
                        npc = true;
                        t = n; //only one NPC can use ILS at the same time
                        this.$N = null;
                        t.awardEquipment("EQ_ILS");
                        bestsd = sd;
                    } else if (nd > near) { //no ILS, just a speedup
                        if (n.injectorSpeedFactor > 0)
                            n.desiredSpeed = n.maxSpeed * n.injectorSpeedFactor;
                        else 
                            n.desiredSpeed = 7 * n.maxSpeed;    
                    }
                }
            }
//            if(t) player.consoleMessage(t.name+" "+t.entityPersonality+" ");//debug
        }
    }

    if (t && t.isValid && (tis || npc)) {
        if (npc) { //NPC ILS
            p = t;
            t = t.target;
            if (debug && p != L) log(this.name, p.displayName+" will use ILS");
        }

        var warning = false;
        if (p != pship && this.$S && this.$S_max_speed && this.$S_max_speed >= 0) {
            if (logging) log(this.name, "Restoring "+t.displayName+" max speed back to "+this.$S_max_speed)
            this.$S.maxSpeed = this.$S_max_speed
            this.$S.maxPitch = this.$S_max_pitch
            this.$S.maxRoll = this.$S_max_roll
            this.$S.maxYaw = this.$S_max_yaw
            this.$S_max_speed = -1
            this.$S_max_pitch = -1
            this.$S_max_roll = -1
            this.$S_max_yaw = -1
        }
        if (this.$S != t || L != p) { //first lock on this target
            this.$S = t; //target is station
            var m = "";
            if (p != pship) {
                m = " of "+p.displayName;
//                if( !player.ship.target ) player.ship.target = p; - confusing, removed
            } else warning = true;
            if (!pship.docked && pship.position.distanceTo(t) <= pship.scannerRange)
                player.consoleMessage("ILS"+m+" set to "+t.displayName, 5);
            if (logging) log(this.name, p.displayName+" is using ILS to dock on "+t.displayName);
            if (p == pship && t.maxSpeed > 0) {
                // "station" is mobile, make sure it stops (Liners!) and does not go into "collision avoidance mode" (Dredgers!)
                this.$S_max_speed = t.maxSpeed
                this.$S_max_pitch = t.maxPitch
                this.$S_max_roll = t.maxRoll
                this.$S_max_yaw = t.maxYaw
                t.maxSpeed = 0
                t.maxPitch = 0
                t.maxRoll = 0
                t.maxYaw = 0
                if (logging) log(this.name, "Setting "+t.displayName+" max speed from "+this.$S_max_speed+" to 0")
            }
        }
        this.$L = L = p; //last ILS user ship
        

        var ph = p.heading;
        pp = p.position; //must set again for NPCs
        if( !this.$P ) this.$P = ph;
        var s = this.$S; //station
        var mp = p.maxPitch;
        var mr = p.maxRoll;
        var my = p.maxYaw;
//        log(this.name, "Station: "+s);//debug
        var u = s.subEntities;
        var i = 0;
        if( u ) while( u[i] && !u[i].isDock ) i++;
        var dp = s.position; //determine dock position
        var dr = s.vectorRight; //save for wide npc
        var dz = s.boundingBox.z;
        var dock, dh;

        if( u && u[i] && u[i].isDock ) {
            dock = u[i]; // de-referencing to make easier to read
            dp = s.position.add(s.heading.multiply( dock.position.z )).
                            add(s.vectorUp.multiply( dock.position.y )).
                            add(dr.multiply( dock.position.x ));
            var x = dock.boundingBox.x, y = dock.boundingBox.y;
            var doo = dock.orientation; //original dock orientation in station coordinates, normal orientation mean no Torus Station
            var dow = dock.orientation.multiply(s.orientation) // final dock orienatation in world coordinates
            dh = dock.heading.rotateBy(dow); // dock heading in final world coordinates
            if( x < y ) {
                maxw = x; //rotated dock model (coriolis, ico)
                /* dybal: this fixes the dock orientation on tetrahedron and octahedron stations: the dock is taller than its width and not rotated relative to the station, so we have to use the stations vectorUp as the dock's wider direction */
                dr = dow.vectorUp();
            } else {
                maxw = y; //dodo, etc.
                dr = dow.vectorRight();
            }
            dz = u[i].boundingBox.z; //save for wide npc
//            if (debug && !npc && counter % 60 === 0) // tracing to understand dock orientation
//                log(this.name, s.displayName+" ("+s.dataKey+"), Dock x,y:"+x.toFixed(1)+","+y.toFixed(1)+", o_orientation:"+[doo.w,doo.x,doo.y,doo.z]+", w_orientation:"+[dow.w,dow.x,dow.y,dow.z]+", relPosition:"+dock.position+", absPosition:"+dp+", Up:"+dow.vectorUp()+", Right:"+dow.vectorRight()+", Head:"+dh+", Station Up:"+s.vectorUp+", Station Right:"+s.vectorRight+", Head:"+s.heading+", direction:"+dr+", position:"+s.position+", distance:"+pp.distanceTo(dp).toFixed(1))
        }
        var p2d = dp.subtract( pp );
        var dd = p2d.magnitude(); //dock distance, used far below for NPCs also
        if (warning) {
            if (p.boundingBox.x < maxw)
                player.consoleMessage("You can dock without roll", 5);
            else if (dd < noinjdist)
                player.consoleMessage("Auto roll is active to fit into the dock", 5);
            else player.consoleMessage("Auto roll will be active before the dock", 5);
        }
        var d = dd / 2;
        var w = p2d.direction();
        var an_pd = dh.angleTo(w); // angle from ship position to docking lane (in the lane if angle near PI since dh is vector exiting the dock)
//        player.consoleMessage(d +" "+ s.heading.angleTo(w)); //debug
        if (!npc && dh.dot(w) >= 0) {
            if (debug && !npc && counter == 0) log(this.name, "angle:"+an_pd.toFixed(3)+"("+(an_pd/halfPI).toFixed(3)+")"+", dot:"+dh.dot(w).toFixed(3)+", NOT engaging ILS");
            // don not engage ILS if the ship is not "in front" of the dock
            return;
        }
        if (debug && !npc && counter == 0) log(this.name, "angle:"+an_pd.toFixed(3)+"("+(an_pd/halfPI).toFixed(3)+")"+", dot:"+dh.dot(w).toFixed(3)+", engaging ILS");
        if (d < 100 || ( an_pd < halfPI && d < 300 )) //fix for rock hermit
            d = -500; //for sure, do not turn back if crossed the point in the dock
        else {
            if (an_pd < 3.14) { //player is not before the dock yet
                var na = PI - an_pd;
                na = na * na;
                d += na * 2000;
                //extend distance if apporaching behind from the station
            }
            d = Math.min( 9500, d ); //max. at the Nav Buoy
        }
        var ap = dp.add(dh.multiply(d)); //aimed position

        var p2ap_dir = ap.subtract(pp).direction();
        if (this.$V && this.$V.isVisualEffect) 
            this.$V.position = ap;
        else 
            this.$V = system.addVisualEffect("ils-marker", ap);
        
        var an_ap = ph.angleTo(p2ap_dir);

        var ato = 0;
        if (ph && this.$P) ato = ph.angleTo(this.$P);//angle to the Previous Heading
        if (debug && !npc && counter == 0) // tracing to check heading convergence to center of docking and smoothness of following marker ball
            log(this.name, "counter:"+counter+", distance:"+d.toFixed(1)+", direction:"+w+", heading:"+p.heading+", angle_to_dock:"+an_pd.toFixed(5)+", aimed:"+ap+", angle_to_ball:"+an_ap.toFixed(5)+", direction_ball:"+p2ap_dir+", previous_heading:"+this.$P+", angle_to_previous:"+ato.toFixed(5));
        /* dybal: changed condition for 'angle' from > 0.0015 to > 0.00015 to get smoother steering */
        if (ato < 0.005 && an_ap > 0.00015
            //steer if no manual steering and not in 0.01 degree
            && p.speed > 1) { //and player is not stopped
            
            var reduc = 1; //reduce pitch in the first 0.5s when start steering
            //if(p.script) { //deprecated, startvector based smoothing is better
                        //        if( !p.script.$ILSTimeSinceSteerStart )
                        //                p.script.$ILSTimeSinceSteerStart = delta;
                        //        else p.script.$ILSTimeSinceSteerStart += delta; //count elapsed time
                        //        reduc = 2*Math.max(0.5, p.script.$ILSTimeSinceSteerStart);
                        //}  //reduc start near 0 and reach 1 after 0.5s
            
            //if the above ato value lower then can not start steering in my intel atom netbook
//            player.consoleMessage(this.$P +" vs "+ ph); //debug

            //limit max turn/step and not too accutate
            var r = Math.min(p.maxPitch*reduc*delta, an_ap/12);

            if (p.script) { //smoother start of ILS steering as the ship has inertia
                var sv = p.script.$ILSSteerStartVect;
                if(sv) 
                    r = Math.min(r, ph.angleTo(sv)/2);
                else {
                    r = 0.1*delta;//need some steering in the first frame
                    p.script.$ILSSteerStartVect = ph; //save steering start vector
                }
            }

            if (!npc && this.$T && this.$T.$TowbarShip && this.$T.$TowbarShip.isValid) {
                var ma = Math.min( 2, p.mass / this.$T.$TowbarShip.mass ); ///small ship max. 2x
                r = r * ma / 3; // 1/3 of the original rate with same mass, min. 1/5 max. 2/3
//                player.consoleMessage("ILS slow steering with towed ship "+ato);//debug
            }
            
            var c = ph.cross(p2ap_dir).direction(); //set the plane where we should rotate in
            var q = p.orientation.rotate( c, -r );
            if (debug && !npc && counter == 0) // tracing to check heading convergence to center of docking
                log(this.name, "Steering: counter:"+counter+", distance:"+d.toFixed(1)+", direction:"+w+", heading:"+ph+", angle_to_dock:"+an_pd.toFixed(5)+", aimed:"+ap+", angle_to_ball:"+an_ap.toFixed(5)+", direction_ball:"+p2ap_dir+", rotate:"+r);
            p.orientation = q;
        } else if (p.script) 
            p.script.$ILSSteerStartVect = null;
                        //p.script.$ILSTimeSinceSteerStart = 0; //reset delta counter

        this.$P = p.heading;
        ph = p.heading; // steering might have altered it

        if (npc) {  //NPC ILS
            var ang = 0, dist = 0, n = null;
            var colldist = 2000; //check for collision within this distance
            var sp = p.maxSpeed;
            var coll = false; //check for possible collision with another ship
            var targets = this.$Targets; //filled by checkScanner() in timer
            if (!targets && !(targets.length > 0)) targets = [];
            targets.push(player.ship);
            for (var i = 0; i < targets.length; i++) {
                n = targets[i];
                if (n && n.isValid) {
                    dist = pp.distanceTo(n);
                    if (dist < colldist) { //is there a ship within 2km?
                        ang = p.heading.angleTo(pp.subtract(n));
                        //if( n == player.ship ) log(this.name, ang);//debug
                        if (ang > 2) { //facing within 65 degree?
                            coll = n; //store the nearest dangerous ship
                            colldist = dist; //used below for desiredSpeed also
                        }
                    }
                }
            }
            if (p.boundingBox.x > maxw && dd < noinjdist) { //wide ship in the last 2km
                p.destination = ap; //AI will slow down near the ap
                if (dd > p.collisionRadius + 120 + dz/2) {
                    p.desiredSpeed = Math.min(sp, 200); //approach slowly
                } else {
                    ang = p.vectorUp.angleTo(dr); //angle between npc and dock
                    var r = 0;
                    if (this.$N || ang > 1.4 && ang < 1.74) { //check upside down also
                        //right angle to start docking
                        this.$N = true;//prevent stop after station rolled
                        p.desiredSpeed = sp;
                        p.destination = ap.add(ap.subtract(pp).multiply(1000));
                        r = this.$S.roll * delta; //roll as the station
                    } else {
                        p.desiredSpeed = 0; //wait for the right angle
                        //need roll especially at non rolling stations
                        r = p.maxRoll/2 * delta;
                        if( this.$S.roll > 0 ) r = -r; //roll opposite
                    }
                    p.orientation = p.orientation.rotate(ph, r);
                    //player.consoleMessage(ang+" "+r);//debug
                }
            } else {
                if (dd > noinjdist) { //use Injector far from the station
                    if (p.injectorSpeedFactor > 0)
                        p.desiredSpeed = p.injectorSpeedFactor * sp;
                    else p.desiredSpeed = 7 * sp; //for Oolite 1.80
                } else p.desiredSpeed = sp; //normal speed near the station
                //place destination farther to do not slow down when ap is near
                p.destination = ap.add(ap.subtract(pp).multiply(5));
                //this.$V.position = p.destination; //debug
            }
            if (coll) { //slow down when another ship is near, as slow as near
                p.desiredSpeed = Math.min(p.desiredSpeed, 0.1*colldist);
                //log(this.name, ang);//debug
            }
                p.desiredRange = 0;
//             player.consoleMessage(p.desiredSpeed+" "+Math.round(p.speed));//debug
        } else { //player rotation before the dock if wide ship in the last 2km
            var an2lane = dh.angleTo(p2ap_dir);
            var maxSpeedRef = this.$maxSpeedRef;
            // how fast we have to get in sync... we can do it slowly if the player's ship is going slow
            var abruptness = (p.speed < maxSpeedRef ? Math.max(0.1, 0.9 * p.speed / maxSpeedRef) : 0.9);

            // (no player steering) && (rotation sync necessary) && (start farther if going fast) && (angle to the docking lane <  60 degrees)
            if (ato < 0.005 && p.boundingBox.x > maxw && dd < 10*abruptness*noinjdist && Math.abs(an2lane - PI) < 1) {
                var clock_wise = (s.roll > 0); // if station rotates anti-clockwise, ship facing opposite way should rotate clockwise
                var ang = p.vectorUp.angleTo(dr); //angle between player ship UP and dock RIGHT, should be PI/2 when in sync
                var ang_diff = ang - halfPI;
                var ang_diff_abs = Math.abs(ang_diff);
                var signal = (clock_wise === true ? 1 : -1); // anti-clockwise angles are negative for rotate method
                var max_r = p.maxRoll * delta; // max angle the ship is able to roll in the interval between FCBs
                var r, r_desired;
                var c;
                if( ang_diff_abs < 0.015 ) {
                    // in sync, just roll with the station
                    c = "A";
                    if (s.roll > 0)
                        r_desired = Math.min(s.roll * delta, max_r); //roll as the station
                    else
                        r_desired = - Math.min(-s.roll * delta, max_r); //roll as the station
                    r = r_desired; // just for the debug message bellow
/* dybal: so far it seems the strategies A and C are enough
//                } else if (dd > dock.boundingBox.z+50 ) { // not in the dock yet
//                    c = "B";
//                    r = p.maxRoll/2 * delta;
//                    if( s.roll > 0 ) r = -r; //roll opposite gets there faster
*/
                } else { 
                    c = "C";
                    if (ang_diff > 0 && clock_wise === true || ang_diff < 0 && clock_wise === false) {
                        // lagging, should rotate the same direction as the station
                        r_desired = s.roll*delta + signal * abruptness * ang_diff_abs;
                    } else {
                        // advanced, should rotate opposite direction as the station
                        r_desired = s.roll*delta - signal * abruptness * ang_diff_abs;
                    }
                    // limit the angle to the ship roll capability
                    if (r_desired > 0)
                        r =  Math.min(r_desired, max_r);
                    else
                        r = - Math.min(-r_desired, max_r);
                }
                if (debug && !npc && counter == 0) // tracing to check heading convergence to center of docking
                    log(this.name, "counter:"+counter+", distance:"+dd.toFixed(0)+"("+d.toFixed(1)+"), angle_to_lane:"+an2lane.toFixed(5)+", angle_to_dock:"+ang.toFixed(3)+"("+ang_diff.toFixed(5)+"), clock-wise:"+clock_wise+", signal:"+signal+", Case:"+c+", delta:"+delta.toFixed(4)+", rotate:"+r.toFixed(5)+", r_desired:"+r_desired.toFixed(4)+", max_r:"+max_r.toFixed(4)+", abrupt:"+abruptness.toFixed(2));
                p.orientation = p.orientation.rotate(ph, r);
                //player.consoleMessage(ang+" "+r);//debug
            } else if (debug && counter == 0) log(this.name, "Rotation sync not necessary");
        }
    } else this.$ClearILS(); //no target or a new target is not station, etc.
//    if (debug) log(this.name, "FCB ends");
}

